/* -*- Mode: C; tab-width: 4; indent-tabs-mode: t; c-basic-offset: 4 -*- */

/*
 * EggNetworkTree GTK+ Widget - egg-network-tree.c
 * 
 * Copyright (C) 2009 Luca Cavalli <luca.cavalli@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * Authors: Luca Cavalli <luca.cavalli@gmail.com>
 */

#include <gio/gio.h>
#include <gtk/gtk.h>

#include "egg-network-tree.h"

enum {
	PROP_NONE
};

enum {
	HOST_SELECTED_SIGNAL,

	LAST_SIGNAL
};

enum {
	ICON_COLUMN,
	HOSTNAME_COLUMN,
	URI_COLUMN,

	N_COLUMNS
};

#define GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj), EGG_TYPE_NETWORK_TREE, EggNetworkTreePrivate))

struct _EggNetworkTreePrivate {
	GMainLoop *main_loop;

	GtkTreeModel *tree_model;
};

static void	egg_network_tree_class_init (EggNetworkTreeClass *klass);
static void	egg_network_tree_init       (EggNetworkTree      *network_tree);
static void egg_network_tree_finalize   (GObject             *object);

static gboolean _populate_network_tree_idle     (EggNetworkTree *network_tree);
static void     _populate_network_tree          (EggNetworkTree *network_tree);
static void     _tree_view_selection_changed_cb (GtkTreeSelection *selection,
												 EggNetworkTree *network_tree);
static void     _mount_enclosing_volume_done_cb (GObject *source,
												 GAsyncResult *result,
												 gpointer user_data);

static guint signals[LAST_SIGNAL] = { 0 };

G_DEFINE_TYPE (EggNetworkTree, egg_network_tree, GTK_TYPE_TREE_VIEW)

static void
egg_network_tree_class_init (EggNetworkTreeClass *klass)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

	g_type_class_add_private (klass, sizeof (EggNetworkTreePrivate));

	gobject_class->finalize = egg_network_tree_finalize;

	signals[HOST_SELECTED_SIGNAL] =
		g_signal_new ("host-selected",
					  G_TYPE_FROM_CLASS (klass),
					  G_SIGNAL_RUN_LAST,
					  G_STRUCT_OFFSET (EggNetworkTreeClass, host_selected),
					  NULL, NULL,
					  g_cclosure_marshal_VOID__STRING,
					  G_TYPE_NONE, 1, G_TYPE_STRING);
}

static void
egg_network_tree_init (EggNetworkTree *network_tree)
{
	EggNetworkTreePrivate *priv;
	GtkCellRenderer *renderer;
	GtkTreeViewColumn *column;
	GtkTreeSelection *selection;

	network_tree->priv = priv = GET_PRIVATE (network_tree);

	priv->main_loop = g_main_loop_new (NULL, FALSE);

	selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (network_tree));
	gtk_tree_selection_set_mode (selection, GTK_SELECTION_SINGLE);
	g_signal_connect (G_OBJECT (selection), "changed",
					  G_CALLBACK (_tree_view_selection_changed_cb),
					  network_tree);

	gtk_tree_view_set_headers_visible (GTK_TREE_VIEW (network_tree), FALSE);

	priv->tree_model = GTK_TREE_MODEL (gtk_tree_store_new (N_COLUMNS,
														   GDK_TYPE_PIXBUF,
														   G_TYPE_STRING,
														   G_TYPE_STRING));
	gtk_tree_view_set_model (GTK_TREE_VIEW (network_tree), priv->tree_model);
	g_object_unref (G_OBJECT (priv->tree_model));

	column = gtk_tree_view_column_new ();

	renderer = gtk_cell_renderer_pixbuf_new ();
	gtk_tree_view_column_pack_start (column, renderer, FALSE);
	gtk_tree_view_column_add_attribute (column, renderer, "pixbuf", ICON_COLUMN);

	renderer = gtk_cell_renderer_text_new ();
	gtk_tree_view_column_pack_start (column, renderer, TRUE);
	gtk_tree_view_column_add_attribute (column, renderer, "text", HOSTNAME_COLUMN);

	gtk_tree_view_append_column (GTK_TREE_VIEW (network_tree), column);

	_populate_network_tree (network_tree);
}

static void
egg_network_tree_finalize (GObject *object)
{
	EggNetworkTreePrivate *priv;

	priv = GET_PRIVATE (EGG_NETWORK_TREE (object));

	if (g_main_loop_is_running (priv->main_loop))
		g_main_loop_quit (priv->main_loop);

	g_main_loop_unref (priv->main_loop);

	G_OBJECT_CLASS (egg_network_tree_parent_class)->finalize (object);
}

/*
 * Passed GFileInfo must have G_FILE_ATTRIBUTE_STANDARD_ICON attribute set.
 */ 
static GdkPixbuf *
_get_pixbuf_from_info (GFileInfo *info)
{
	GIcon *icon;
	GdkPixbuf *pixbuf = NULL;

	g_return_val_if_fail (info, NULL);

	icon = g_file_info_get_icon (info);

	if (G_IS_THEMED_ICON (icon)) {
		gchar const * const *names;

		names = g_themed_icon_get_names (G_THEMED_ICON (icon));
		pixbuf = gtk_icon_theme_load_icon (gtk_icon_theme_get_default (), *names, 24, 0, NULL);
	}
	else if (G_IS_FILE_ICON (icon)) {
		GFile *icon_file;
		gchar *path;

		icon_file = g_file_icon_get_file (G_FILE_ICON (icon));
		path = g_file_get_path (icon_file);
		pixbuf = gdk_pixbuf_new_from_file_at_size (path, 24, 24, NULL );
		g_free (path);
		g_object_unref (G_OBJECT (icon_file));
	}
	g_object_unref (G_OBJECT (icon));

	return (pixbuf);
}

static gboolean
_populate_network_tree_idle (EggNetworkTree *network_tree)
{
	EggNetworkTreePrivate *priv;
	GFile *root;
	GFileEnumerator *group_enum;
	GFileInfo *group_info;
	GtkTreeIter iter;
	GtkTreeStore *tree_store;
	GError *error;

	priv = GET_PRIVATE (network_tree);

	root = g_file_new_for_uri ("smb:///");

	g_file_mount_enclosing_volume (root, G_MOUNT_MOUNT_NONE, NULL, NULL,
								   _mount_enclosing_volume_done_cb,
								   (gpointer) network_tree);
	g_main_loop_run (priv->main_loop);

	error = NULL;
	group_enum = g_file_enumerate_children (root,
											G_FILE_ATTRIBUTE_STANDARD_DISPLAY_NAME ","
											G_FILE_ATTRIBUTE_STANDARD_TARGET_URI ","
											G_FILE_ATTRIBUTE_STANDARD_ICON,
											0, NULL, &error);

	if (group_enum == NULL) {
 		g_warning ("Error enumerating %s children - %s", "smb:///", error->message);
		g_error_free (error);

		g_object_unref (root);

		return FALSE;
	}

	tree_store = GTK_TREE_STORE (priv->tree_model);

	error = NULL;
	while ((group_info = g_file_enumerator_next_file (group_enum, NULL, &error)) != NULL) {
		const char *group_display_name;
		char *group_uri;
		GdkPixbuf *group_pixbuf = NULL;
		GFile *group;
		GFileEnumerator *host_enum;
		GFileInfo *host_info;
		GtkTreeIter child;

		if (error != NULL) {
			g_warning ("Error enumerating %s children - %s", "smb:///", error->message);
			g_error_free (error);

			g_object_unref (group_info);

			error = NULL;
			continue;
		}

		group_display_name = g_file_info_get_display_name (group_info);
		group_uri = g_file_info_get_attribute_as_string (group_info,
														 G_FILE_ATTRIBUTE_STANDARD_TARGET_URI);

		group_pixbuf = _get_pixbuf_from_info (group_info);

		gtk_tree_store_append (tree_store, &iter, NULL);
		gtk_tree_store_set (tree_store, &iter,
							ICON_COLUMN, group_pixbuf,
							HOSTNAME_COLUMN, group_display_name,
							URI_COLUMN, NULL,
							-1);

		group = g_file_new_for_uri (group_uri);

		g_file_mount_enclosing_volume (group, G_MOUNT_MOUNT_NONE, NULL, NULL,
									   _mount_enclosing_volume_done_cb,
									   (gpointer) network_tree);
		g_main_loop_run (priv->main_loop);

		host_enum = g_file_enumerate_children (group,
											   G_FILE_ATTRIBUTE_STANDARD_DISPLAY_NAME ","
											   G_FILE_ATTRIBUTE_STANDARD_TARGET_URI ","
											   G_FILE_ATTRIBUTE_STANDARD_ICON,
											   0, NULL, &error);

		if (host_enum == NULL) {
			g_warning ("Error enumerating %s children - %s", group_uri, error->message);
			g_error_free (error);

			g_object_unref (group);

			error = NULL;
			continue;
		}

		error = NULL;
		while ((host_info = g_file_enumerator_next_file (host_enum, NULL, &error)) != NULL) {
			const char *host_display_name;
			char *host_uri;
			GdkPixbuf *host_pixbuf = NULL;

			if (error != NULL) {
				g_warning ("Error enumerating %s children - %s", group_uri, error->message);
				g_error_free (error);

				g_object_unref (host_info);

				error = NULL;
				continue;
			}

			host_display_name = g_file_info_get_display_name (host_info);
			host_uri = g_file_info_get_attribute_as_string (host_info,
															G_FILE_ATTRIBUTE_STANDARD_TARGET_URI);

			host_pixbuf = _get_pixbuf_from_info (host_info);

			gtk_tree_store_append (tree_store, &child, &iter);
			gtk_tree_store_set (tree_store, &child,
								ICON_COLUMN, host_pixbuf,
								HOSTNAME_COLUMN, host_display_name,
								URI_COLUMN, host_uri,
								-1);

			g_free (host_uri);
			g_object_unref (host_info);
			g_object_unref (host_pixbuf);
		}

		g_free (group_uri);
		g_object_unref (group);
		g_object_unref (group_info);
		g_object_unref (group_pixbuf);
	}

	g_object_unref (group_enum);
	g_object_unref (root);

	return FALSE;
}

static void
_populate_network_tree (EggNetworkTree *network_tree)
{
	g_idle_add ((GSourceFunc) _populate_network_tree_idle, network_tree);
}


static void
_tree_view_selection_changed_cb (GtkTreeSelection *selection, EggNetworkTree *network_tree)
{
	GtkTreeIter iter;
	GtkTreeModel *tree_model;
	gchar *uri;

	if (gtk_tree_selection_get_selected (selection, &tree_model, &iter)) {
		gtk_tree_model_get (tree_model, &iter, URI_COLUMN, &uri, -1);
		g_signal_emit_by_name (G_OBJECT (network_tree), "host-selected", uri);
		g_free (uri);
	}
}

static void
_mount_enclosing_volume_done_cb (GObject *source, GAsyncResult *result, gpointer user_data)
{
	EggNetworkTreePrivate *priv;

	priv = GET_PRIVATE (EGG_NETWORK_TREE (user_data));

	g_file_mount_enclosing_volume_finish (G_FILE (source), result, NULL);

	g_main_loop_quit (priv->main_loop);
}

GtkWidget *
egg_network_tree_new (void)
{
	EggNetworkTree *network_tree;

	network_tree = g_object_new (EGG_TYPE_NETWORK_TREE, NULL);

	return GTK_WIDGET (network_tree);
}

void
egg_network_tree_refresh (EggNetworkTree *network_tree)
{
	EggNetworkTreePrivate *priv;

	g_return_if_fail (network_tree != NULL);

	priv = GET_PRIVATE (network_tree);

	gtk_tree_store_clear (GTK_TREE_STORE (priv->tree_model));
	_populate_network_tree (network_tree);

	g_signal_emit_by_name (G_OBJECT (network_tree), "host-selected", NULL);
}
